#include "Log.h"
#include "arch/CPU.h"
#include "arch/MMU.h"
#include "arch/Serial.h"
#include "arch/Timer.h"
#include "boot/Init.h"
#include "config.h"
#include "memory/Heap.h"
#include "memory/KernelVM.h"
#include "memory/MemoryManager.h"
#include "thread/Scheduler.h"
#include <luna/Result.h>
#include <luna/Units.h>

void async_thread()
{
    while (true)
    {
        Thread* current = Scheduler::current();
        kinfoln("Ticks: %lu, %lu user, %lu kernel, %lu idle", current->ticks, current->ticks_in_user,
                current->ticks_in_kernel, Scheduler::idle()->ticks);

        CPU::wait_for_interrupt();

        kernel_sleep(1000);
    }
}

void heap_thread()
{
    CPU::disable_interrupts();
    dump_heap_usage();
    kdbgln("uses %lu vm pages", KernelVM::used() / ARCH_PAGE_SIZE);
    while (true) kernel_sleep(UINT64_MAX);
}

Result<void> init()
{
    kinfoln("Starting Moon %s", MOON_VERSION);

    kinfoln("Current platform: %s", CPU::platform_string());

    kinfoln("Current processor: %s", CPU::identify().value_or("(unknown)"));

    Timer::init();

    kinfoln("Total memory:    %s", to_dynamic_unit(MemoryManager::total()).release_value().chars());
    kinfoln("Free memory:     %s", to_dynamic_unit(MemoryManager::free()).release_value().chars());
    kinfoln("Used memory:     %s", to_dynamic_unit(MemoryManager::used()).release_value().chars());
    kinfoln("Reserved memory: %s", to_dynamic_unit(MemoryManager::reserved()).release_value().chars());

    Thread::init();
    Scheduler::init();

    TRY(Scheduler::new_kernel_thread(async_thread));
    TRY(Scheduler::new_kernel_thread(heap_thread));

    CPU::platform_finish_init();

    CPU::enable_interrupts();

    return {};
}

extern "C" [[noreturn]] void _start()
{
    Init::check_magic();
    Init::early_init();
    auto rc = init();
    if (rc.has_error()) kerrorln("Runtime error: %s", rc.error_string());
    CPU::idle_loop();
}